clear all;
clc;
close all;
% This script conducts a simulation of the notched-noise masking experiment
% using the qAF toolbox. The experiment is configured in the file
% "exp_config_roex.m", and the control of the experimental protocol is in
% this script. In principle, the class definition file "qAF.m" do not need
% to be changed.

% For more information about the qAF procedure, see:
% Shen, Y., Sivakumar, R., & Richards, V. M. (2014). Rapid estimation of 
% high-parameter auditory-filter shapes. The Journal of the Acoustical 
% Society of America, 136(4), 1857-1868.
%
% Shen, Y., & Richards, V. M. (2013). Bayesian adaptive estimation of the 
% auditory filter. The Journal of the Acoustical Society of America, 
% 134(2), 1134-1145.

% It is highly recommended that each qAF track to be preceeded by a warm-up
% period, in which the masker is fixed and the target is initially 
% presented at high sensation levels and then decreased according to an 
% up-down procedure. See Shen et al. (2014) for details.

% This software package is developed at the University of California,
% Irvine and at Indiana University Bloomington by Yi Shen. In case of any
% questions, feel free to email Yi Shen at shen2@indiana.edu. The 
% development of this software package is ongoing, your feedback and
% comments are very welcome.

% Virtual listener. The parameters are arranged as pu, pl, t, 10logw, and
% 10logK
phi0 = [44.7281   45.8076    8.0470  -41.3365   10.8181];

N = 80;     % Total number of trials
% Number of parameters. If set to 5, it is the standards roex(pwpt) filter.
% If set to 4, the parameter t is fixed to a value of 9. If set to 3, t is
% set to 9 and pu and pl are assumed to be identical.
npar = length(phi0);
% Configure the experiment and store all relevent settings to a structure.
configpar = exp_config_roex(npar, 50, 2000);
% Constructe a qAF object using the structure generated by exp_config_roex.
qaf = qAF(configpar);
% For simulations, set the parameters of the virtual listener.
qaf.setPhi0(phi0);

figure(1);
for i = 1:N
    
    % Simulate a trial. Note that qaf.xnext stores the optimal stimulus for
    % the next trial. The response variable r is defined in terms of
    % correctness, with r=1 and 0 indicate correct and incorrect,
    % respectively.
    r = qaf.simulateResponse(qaf.xnext);
    
    %------------------------------------
    % For an actual psychophysical experiment, a user-defined function is
    % needed here, in the format of:
    %
    % r = UserFunction(qaf.xnext);
    %
    % It takes the stimulus parameter as the input argument and returns,
    % presents the stimulus, obtains the response, and return the response
    % correctness as the output argument.
    %------------------------------------
    
    % Update model parameters
    qaf.update(r); 
    
    % Plot the stimuli sampled and the auditory filter estimated up to the
    % current trial.
    figure(1);
    delete(get(gcf,'Children'));
    qaf.plotStim();
    drawnow;
end

% Plot the parameter estimates as functions of trial number.
figure(2);
qaf.plotPhi();

%%
% Other methods that are included in the qAF class include:

% Return the ERB estimates across all trials.
ERB = qaf.getERB;   

% Predict thresholds using the current estimate of the auditory filter
predict_gu = 0.4;   % Upper notchwidth
predict_gl = 0.5;   % Lower notchwidth
predict_pc = 0.75;  % Proportion correct at which threshold is defined
Thre = qaf.predictThre([predict_gu predict_gl], predict_pc); 

% Plot the posterior distribution
figure(3);
qaf.plotP();



