classdef qAF <handle
    
    properties
        %basic properties
        x   % stimulus parameter
        xnext % the next stimulus based on the previous data;
        phi0 = []; %the parameters to be estimated in the weibull psychometric function
        phi %the mean of the posterior parameter distribution
        P   %the covariance matrix of the posterior parameter distribution
        r   %the listener's responses (correctness)
        PC  %percent correct
        par %the data structure containing the configurations of the parameter space   
        n=0;   %the trial number
        
        % function handles for the psychometric function and the jacobian
        % matrix
        psycfun
        jacobian
        
        % variable names reserved for user defined data
        userdata01
        userdata02
        userdata03
        userdata04
    end
    
    methods
        
        %constructor
        function qaf = qAF(par)
            reset(qaf, par);
        end  
        
        % INITIALIZATION
        function reset(qaf, par)
            qaf.par = par;
            qaf.setP0();
            qaf.x;
            qaf.r;
            qaf.n = 0;
            
            % labels for the stimulus parameters for plotting purposes
            qaf.par.x_label = ...
                {'Normalized upper notchwidth','Normalized lower notchwidth','Masker level (dB SPL)'};  
            
            %initlize the parameters in updown algorithm
            if strcmp(par.model,'roex')
                if par.npar == 3
                    qaf.psycfun = @roex3_psycfun;
                    qaf.jacobian = @roex3_jacobian;
                    % limits and name for each model parameter for plotting purposes
                    qaf.par.phi_lim = [10 -60 -10; 70 0 20];
                    qaf.par.phi_label = {'p','10log10(w)','10log10(K)'};
                    qaf.par.diffusion = 0.05;
                elseif par.npar == 4
                    qaf.psycfun = @roex4_psycfun;
                    qaf.jacobian = @roex4_jacobian;
                    % limits and name for each model parameter for plotting purposes
                    qaf.par.phi_lim = [10 10 -60 -10; 70 70 0 20];
                    qaf.par.phi_label = {'pu','pl','10log10(w)','10log10(K)'};
                    qaf.par.diffusion = 0.02;
                elseif par.npar == 5
                    qaf.psycfun = @roex5_psycfun;
                    qaf.jacobian = @roex5_jacobian;
                    % limits and name for each model parameter for plotting purposes
                    qaf.par.phi_lim = [10 10 0 -60 -10; 70 70 20 0 20];
                    qaf.par.phi_label = {'p_u','p_l','t','10log10(w)','10log10(K)'};
                    qaf.par.diffusion = 0.01;
                end
            else
                error('Please set "model.par" to "roex".');
            end
            
            qaf.xnext = findx_min_entropy(qaf.phi,qaf.P,qaf.par.x_lim,qaf.par.x_n,...
                @qaf.psycfun, @qaf.jacobian, qaf.par);        
        end
        
        %set the hypothesis made on the parameters to be estimated
        function setP0(qaf)
            %set the prior for phi
            if length(qaf.par.prior_phi) == qaf.par.npar &&...
                    length(qaf.par.prior_P) == qaf.par.npar
                qaf.phi = qaf.par.prior_phi;
                qaf.P = qaf.par.prior_P;
            else
                error('The size of the prior parameters should agree with the choice of the model.');
            end
        end
        
        % ITERATION
        
        %Update posterior and xnext based on the previous posterior and the
        %new response r.
        function update(qaf, r) 
            qaf.n = qaf.n + 1;
            qaf.x(qaf.n,:) = qaf.xnext;
            qaf.r(qaf.n,:) = r;
            
            % update model parameters
            [qaf.phi(:,qaf.n+1), qaf.P(:,:,qaf.n+1)] =...
                Kalman_update(qaf.phi(:,qaf.n),qaf.P(:,:,qaf.n), qaf.x(qaf.n,:), qaf.r(qaf.n,:),...
                @qaf.psycfun, @qaf.jacobian, qaf.par);
            
            %Find the next signal strength
            qaf.xnext = ...
                findx_min_entropy(qaf.phi(:,qaf.n+1),qaf.P(:,:,qaf.n+1),qaf.par.x_lim,qaf.par.x_n,...
                @qaf.psycfun, @qaf.jacobian, qaf.par);
            
            % update percent correct
            qaf.PC = sum(qaf.r)/length(qaf.r);
        end
        
        % SIMULATION
        
        %set the parameters in the psychometric function of the virtual
        % observer for simulation
        function setPhi0(qaf, phi0)
            if length(phi0) == qaf.par.npar
                qaf.phi0 = phi0;
            else
                error('The number of parameters for the virual listener should match that of the model.');
            end
        end
        
        %given a stimulus, get a simulated virtual response
        function r = simulateResponse(qaf, x)
            %given a stimulus, get a response
            r = rand<qaf.psycfun(x, qaf.phi0, qaf.par);
        end
        
        % RETRIEVAL AND PLOTTING
        
        % return the ERB estimate
        function [ERB, ERB_std] = getERB(qaf)
            ntrials = size(qaf.phi,2);
            ERB = zeros(1,ntrials);
            ERB_std = ERB;
            nMCsmps = 1000;
            if strcmp(qaf.par.model,'roex')
                for itrial = 1:ntrials
                    if qaf.par.npar == 4 || qaf.par.npar == 5
                        ERB(itrial) = (2./qaf.phi(1,itrial)+2./qaf.phi(2,itrial))*qaf.par.stim_par.fo;
                        pu_smps = qaf.phi(1,itrial)+sqrt(qaf.P(1,1,itrial))*randn(nMCsmps,1);
                        pl_smps = qaf.phi(2,itrial)+sqrt(qaf.P(2,2,itrial))*randn(nMCsmps,1);
                        ERB_smps = (2./pu_smps+2./pl_smps)*qaf.par.stim_par.fo;
                        ERB_std(itrial) = std(ERB_smps);
                    elseif qaf.par.npar == 3
                        ERB(itrial) = (4./qaf.phi(1,itrial))*qaf.par.stim_par.fo;
                        p_smps = qaf.phi(1,itrial)+sqrt(qaf.P(1,1,itrial))*randn(nMCsmps,1);
                        ERB_smps = (4./p_smps)*qaf.par.stim_par.fo;
                        ERB_std(itrial) = std(ERB_smps);
                    end
                end
            end
        end
        
        % plot the parameter estimates
        function plotPhi(qaf)
            if strcmp(qaf.par.model,'roex')
                if qaf.par.npar > 3
                    subplot_idx = [2 3];
                else
                    subplot_idx = [2 2];
                end
                for ipar = 1:qaf.par.npar
                    subplot(subplot_idx(1), subplot_idx(2), ipar);
                    phi_std = sqrt(squeeze(qaf.P(ipar,ipar,:)));
                    phi_tmp = [qaf.phi(ipar,:)'-phi_std, qaf.phi(ipar,:)', qaf.phi(ipar,:)'+phi_std];
                    h = plot(phi_tmp,'k');
                    set(h(1),'LineStyle','--');
                    set(h(2),'LineWidth',2);
                    set(h(3),'LineStyle','--');
                    xlabel('Trial number');
                    ylabel(qaf.par.phi_label{ipar});
                    xlim([0 qaf.n]);
                    ylim(qaf.par.phi_lim(:,ipar));
                end

                % plot ERB estimates
                subplot(subplot_idx(1), subplot_idx(2), ipar+1);
                [ERB, ERBstd] = qaf.getERB();
                h = plot([ERB'-ERBstd' ERB' ERB'+ERBstd'],'k');
                set(h(1),'LineStyle','--');
                set(h(2),'LineWidth',2);
                set(h(3),'LineStyle','--');
                xlabel('Trial number');
                ylabel('ERB (Hz)');
                xlim([0 qaf.n]);
                if qaf.par.npar == 4 || qaf.par.npar == 5
                    ylim(sort(sum(2./qaf.par.phi_lim(:,1:2),2))*qaf.par.stim_par.fo);
                elseif qaf.par.npar == 3
                    ylim(sort(4./qaf.par.phi_lim(:,1))*qaf.par.stim_par.fo);
                end
            end
        end
        
        % plot the posterior parameter distribution
        function plotP(qaf,varargin)
            set(gcf,'units','normalized','position',[0 0 1 1]);
            
            axislabel = qaf.par.phi_label;
            if nargin >=2
                gradation = varargin{1};
            else
                gradation = 25;
            end
            ndim = qaf.par.npar;
            x_lim = qaf.par.phi_lim;
            mu = qaf.phi(:,end)';
            Sigma = qaf.P(:,:,end);

            for i = 1:ndim
                for j = 1:ndim
                    subplot(ndim,ndim,(i-1)*ndim+j);
                    if i==j
                        xplot = linspace(x_lim(1,i),x_lim(2,i),gradation);
                        plot(xplot,normpdf(xplot,mu(i),Sigma(i,j)));
                        xlim(x_lim(:,i)');
                        if ~isempty(axislabel)
                            xlabel(axislabel{i});
                        end
                        ylabel('Probability density');
                    else
                        iso_density_contour(mu([j i]),Sigma([j i],[j i]),x_lim(:,[j i]),gradation);
                        if ~isempty(axislabel)
                            xlabel(axislabel{j});
                            ylabel(axislabel{i});
                        end
                    end
                end
            end
        end
        
        % plot the auditory filter
        function plotAF(qaf)
            phi_tmp = qaf.phi(:,end);
            f = linspace(qaf.par.stim_par.fo/4,qaf.par.stim_par.fo*2,200)';
            if strcmp(qaf.par.model,'roex')
                af = @(p, g) (1 + p.*g).*exp(-p.*g);
                g = (f-qaf.par.stim_par.fo)/qaf.par.stim_par.fo;
                auditory_filter_shape = zeros(size(f));
                if qaf.par.npar == 3
                    pu = phi_tmp(1);
                    pl = phi_tmp(1);
                    t = 9;
                    w = 10.^(phi_tmp(2)/10);
                    auditory_filter_shape(g<0) = (1-w)*af(pl,-g(g<0))+w*af(t,-g(g<0));
                    auditory_filter_shape(g>=0) = af(pu,g(g>=0));
                    
                elseif qaf.par.npar == 4
                    pu = phi_tmp(1);
                    pl = phi_tmp(2);
                    t = 9;
                    w = 10.^(phi_tmp(3)/10);
                    auditory_filter_shape(g<0) = (1-w)*af(pl,-g(g<0))+w*af(t,-g(g<0));
                    auditory_filter_shape(g>=0) = af(pu,g(g>=0));
                    
                elseif qaf.par.npar == 5
                    pu = phi_tmp(1);
                    pl = phi_tmp(2);
                    t = phi_tmp(3);
                    w = 10.^(phi_tmp(4)/10);
                    auditory_filter_shape(g<0) = (1-w)*af(pl,-g(g<0))+w*af(t,-g(g<0));
                    auditory_filter_shape(g>=0) = af(pu,g(g>=0));
                end
                auditory_filter_shape = 10*log10(auditory_filter_shape);
            end
            semilogx(f,auditory_filter_shape,'k','LineWidth',2);
            xlabel('Frequency (Hz)');
            ylabel('Magnitude response (dB)');
            axis([qaf.par.stim_par.fo/4 qaf.par.stim_par.fo*2 -50 10]);
            set(gca,'XTick',qaf.par.stim_par.fo*[0.25 0.32 0.4 0.5 0.625 0.8 1 1.25 1.6 2],...
                'XMinorTick','off');
            ERB = qaf.getERB();
            ERB = ERB(end);
            text(qaf.par.stim_par.fo*0.8, -40, ['ERB = ' num2str(ERB,'%1.1f') ' Hz']);
        end
        
        % plot the stimulus
        function plotStim(qaf)
            subplot(211);
            fu = qaf.par.stim_par.fo+qaf.x(:,1)*qaf.par.stim_par.fo;
            fl = qaf.par.stim_par.fo-qaf.x(:,2)*qaf.par.stim_par.fo;
            scatter(fu,qaf.x(:,3));hold on;
            scatter(fl,qaf.x(:,3));hold off;
            xlabel('Notch position (Hz)');
            ylabel('Masker Level (dB SPL)');
            xlim([qaf.par.stim_par.fo/4 qaf.par.stim_par.fo*2]);
            ylim(qaf.par.x_lim(:,3));
            set(gca,'XTick',qaf.par.stim_par.fo*[0.25 0.32 0.4 0.5 0.625 0.8 1 1.25 1.6 2],...
                'XMinorTick','off','XScale','log');

            ax1 = gca;
            ax2 = axes('YAxisLocation','right','Position',get(ax1,'Position'));
            set(gcf,'CurrentAxes',ax2);
            qaf.plotAF;
            box off;
            xlim([qaf.par.stim_par.fo/4 qaf.par.stim_par.fo*2]);
            set(gca,'XTick',[],'XTickLabel',[],...
                'XMinorTick','off','XScale','log', 'YDir','reverse');
%             set(ax2,'YAxisLocation','right','Color','none','Position',get(ax1,'Position'));
            xlabel([]);

            subplot(223)
            plot(qaf.x(:,1:2));
            xlabel('Trial number');
            ylabel('Normalized Notchwidth');
            legend('upper','lower');
            xlim([0, size(qaf.x,1)]);

            subplot(224)
            plot(qaf.x(:,3));
            xlabel('Trial number');
            ylabel('Masker Level (dB SPL)');
            xlim([0, size(qaf.x,1)]);
            
            set(ax2,'YAxisLocation','right','Color','none','Position',get(ax1,'Position'));
        end
        
        % predict thresholds
        function thre = predictThre(qaf,notchwidth, pc)
            if size(notchwidth,2) ~= 2
                error('Please arrange the upper and lower notchwidths in two columns.');
            end
            nconds = size(notchwidth,1);
            thre = zeros(nconds,1);
            for icond = 1:nconds
                [dummy, thre0] = qaf.psycfun([notchwidth(icond,:), 0], qaf.phi(:,end), qaf.par);
                thre(icond) = fminsearch(@(y) (qaf.psycfun([notchwidth(icond,:), y], qaf.phi(:,end), qaf.par) - pc)^2, thre0);
            end
        end
    end
end

%%
% functions for extended kalman filtering and min-entropy research for the
% stimulus.

function [phi, P, l] = Kalman_update(phi, P, x, r, psyc_fun, qAF_Jacobian, pf_par)
    
    % inherent noise source for the random-walk of the parameters
    Q = pf_par.diffusion*diag(diag(P));
    P = P+Q;

    % Expected response and response variability
    mu = psyc_fun(x,phi, pf_par);
    var = mu*(1-mu);
    
    % Compute Kalman gain factor:
    H1 = qAF_Jacobian(x,phi, pf_par);
    K = (P*H1')/(H1*P*H1' + var);

    % Correction based on observation:
    phi_next = phi + K*(r-mu);
    if all(phi_next'>pf_par.phi_lim(1,:)) && all(phi_next'<pf_par.phi_lim(2,:)) 
        phi = phi_next;
        P = P - K*H1*P;
    end
    
    % liklihood
    if r == 1
        l = mu;
    elseif r == 0
        l = 1-mu;
    end
    
end

function xnext = findx_min_entropy(phi,P,x_lim, x_n, psyc_fun, qAF_Jacobian, pf_par)

    % only gridsearch
        gu = linspace(x_lim(1,1),x_lim(2,1),x_n(1));
        gl = linspace(x_lim(1,2),x_lim(2,2),x_n(2));
        siglev = linspace(x_lim(1,3),x_lim(2,3),x_n(3));
        entropy = zeros(length(gu),length(gl),length(siglev));
        for igu = 1:length(gu)
            for igl = 1:length(gl)
                for ilev = 1:length(siglev)
                    entropy(igu,igl,ilev) = find_expected_entropy(phi, P,[gu(igu), gl(igl), siglev(ilev)], psyc_fun, qAF_Jacobian, pf_par);
                end
            end
        end
        [dummy,idx] = min3d(entropy);
        xnext = [gu(idx(1)), gl(idx(2)), siglev(idx(3))];

    % % only fminsearch
    %     xnext = fminsearch(@(x) find_expected_entropy(phi, P,x, psyc_fun, qAF_Jacobian, pf_par)+ ...
    %         (x(1)<x_lim(1,1) | x(1)>x_lim(2,1) | x(2) < x_lim(1,2) | x(2) > x_lim(2,2) | x(3) < x_lim(1,3) | x(3) > x_lim(2,3))*1e6,...
    %         [max(x_lim(:,1:2)) 35],optimset('MaxFunEvals',400,'Display','off'));

%     % combine gridsearch and fminsearch
%     % divide the siglev range into four bins
%     sigbins = 3;
%     siglev = linspace(x_lim(1,3),x_lim(2,3),sigbins+1);
%     siglev = (siglev(1:sigbins)+siglev(2:sigbins+1))/2;
%     xnext_tot = zeros(length(siglev),3);
%     entropy = zeros(length(siglev),1);
%     for ilev = 1:length(siglev)
%         [xnext_tot(ilev,:), entropy(ilev)] = fminsearch(@(x) find_expected_entropy(phi, P,x, psyc_fun, qAF_Jacobian, pf_par)+ ...
%             (x(1)<x_lim(1,1) | x(1)>x_lim(2,1) | x(2) < x_lim(1,2) | x(2) > x_lim(2,2) | x(3) < x_lim(1,3) | x(3) > x_lim(2,3))*1e6,...
%             [mean(x_lim(:,1:2)) siglev(ilev)],optimset('MaxFunEvals',100,'Display','off'));
%     end
% %     idx0 = find(~isnan(entropy));
% %     entropy = entropy(idx0);
% %     xnext_tot = xnext_tot(idx0,:);
%     [dummy,idx] = min(entropy);
%     xnext = xnext_tot(idx,:);
end

function Hexp = find_expected_entropy(phi, P, x, psyc_fun, qAF_Jacobian, pf_par)
    
    H_mulvar_norm = @(phi,P) (length(phi))/2+(length(phi))*log(2*pi)/2+log(det(P))/2;

    [phi1, P1, p1] = Kalman_update(phi, P, x, 1, psyc_fun, qAF_Jacobian, pf_par);
    [phi0, P0, p0] = Kalman_update(phi, P, x, 0, psyc_fun, qAF_Jacobian, pf_par);
    Hexp = p1*H_mulvar_norm(phi1,P1) + p0*H_mulvar_norm(phi0,P0);

end

function [y, idx] = min3d(x)
    if length(size(x))~=3
        error('The input has to be a three-dimensional matrix!');
    end

    [m, n, k] = size(x);

    x_array = reshape(x,m*n*k,1);
    y = min(x_array);
    idx1 = find(x_array==y);
    idx1 = idx1(randperm(length(idx1)));
    idx0 = idx1(1);

    idx = zeros(1,3);
    idx(3) = ceil(idx0/(m*n));
    idx0 = idx0-(idx(3)-1)*m*n;

    idx(2) = ceil(idx0/m);
    idx(1) = rem(idx0,m);
    if idx(1) ==0
        idx(1) = m;
    end
end


%%
% definitions for the psychometric functions and their associated jacobian
% matrices

function [p, N0] = roex3_psycfun(x, phi, par)

    % stimulus parameters in x
    gu = x(1);
    gl = x(2);
    msklev = x(3);

    % model parameters in phi
    pu = phi(1);
    pl = phi(1);
    t = 9;
    w = phi(2);
    K = phi(3);

    % other stimulus parameters
%     gamma = 1/3;
%     beta = 1;
%     f = 2000;
%     Ps = 30;
%     BW = 800;
    gamma = par.pf_par.gamma;
    beta = par.pf_par.beta;
    f = par.stim_par.fo;
    Ps = par.stim_par.siglev;
    BW = par.stim_par.BW;

    fl = (1-gl)*f-BW;
    fu = (1+gu)*f+BW;
    maxgl = (f-fl)/f;
    maxgu = (fu-f)/f;

    roex_integral = @(p,g) -(1/p)*(2+p*g)*exp(-p*g);

    term_lowertip = roex_integral(pl,maxgl)-roex_integral(pl,gl);
    term_lowertail = roex_integral(t,maxgl)-roex_integral(t,gl);
    term_uppertip = roex_integral(pu,maxgu)-roex_integral(pu,gu);

    term1 = f*((1-10^(w/10))*term_lowertip+(10^(w/10))*term_lowertail+term_uppertip);
    N0 = Ps - 10*log10(term1) - K;
    p = gamma+(1-gamma)*(1+exp((msklev-N0)*beta)).^(-1);

end

function H = roex3_jacobian(x,phi, par)
    % all derivative here are calculated numerically
    p = phi(1);
    logw = phi(2);
    logK = phi(3);
    h = 1e-6;   % this determines the local truncation error

    % 1-point numerical derivative (less accurate but fast)
    % ref
    ref = roex3_psycfun(x,[p,logw,logK], par);
    % dpc/dpu
    H(1) = (roex3_psycfun(x,[p+h,logw,logK], par)-ref)/h;
    % dpc/dlogw
    H(2) = (roex3_psycfun(x,[p,logw+h,logK], par)-ref)/h;
    % dpc/dlogK
    H(3) = (roex3_psycfun(x,[p,logw,logK+h], par)-ref)/h;

end

function [p, N0] = roex4_psycfun(x, phi, par)

    % stimulus parameters in x
    gu = x(1);
    gl = x(2);
    msklev = x(3);

    % model parameters in phi
    pu = phi(1);
    pl = phi(2);
    t = 9;
    w = phi(3);
    K = phi(4);

    % other stimulus parameters
%     gamma = 1/3;
%     beta = 1;
%     f = 2000;
%     Ps = 30;
%     BW = 800;
    gamma = par.pf_par.gamma;
    beta = par.pf_par.beta;
    f = par.stim_par.fo;
    Ps = par.stim_par.siglev;
    BW = par.stim_par.BW;

    fl = (1-gl)*f-BW;
    fu = (1+gu)*f+BW;
    maxgl = (f-fl)/f;
    maxgu = (fu-f)/f;

    roex_integral = @(p,g) -(1/p)*(2+p*g)*exp(-p*g);

    term_lowertip = roex_integral(pl,maxgl)-roex_integral(pl,gl);
    term_lowertail = roex_integral(t,maxgl)-roex_integral(t,gl);
    term_uppertip = roex_integral(pu,maxgu)-roex_integral(pu,gu);

    term1 = f*((1-10^(w/10))*term_lowertip+(10^(w/10))*term_lowertail+term_uppertip);
    N0 = Ps - 10*log10(term1) - K;
    p = gamma+(1-gamma)*(1+exp((msklev-N0)*beta)).^(-1);

end

function H = roex4_jacobian(x,phi, par)
    % all derivative here are calculated numerically
    % model parameters in phi
    pu = phi(1);
    pl = phi(2);
    logw = phi(3);
    logK = phi(4);
    h = 1e-6;   % this determines the local truncation error

    % 1-point numerical derivative (less accurate but fast)
    % ref
    ref = roex4_psycfun(x,[pu,pl,logw,logK], par);
    % dpc/dpu
    H(1) = (roex4_psycfun(x,[pu+h,pl,logw,logK], par)-ref)/h;
    % dpc/dpl
    H(2) = (roex4_psycfun(x,[pu,pl+h,logw,logK], par)-ref)/h;
    % dpc/dlogw
    H(3) = (roex4_psycfun(x,[pu,pl,logw+h,logK], par)-ref)/h;
    % dpc/dlogK
    H(4) = (roex4_psycfun(x,[pu,pl,logw,logK+h], par)-ref)/h;
end

function [p, N0] = roex5_psycfun(x, phi, par)

    % stimulus parameters in x
    gu = x(1);
    gl = x(2);
    msklev = x(3);

    % model parameters in phi
    pu = phi(1);
    pl = phi(2);
    t = phi(3);
    w = phi(4);
    K = phi(5);

    % other stimulus parameters
%     gamma = 1/3;
%     beta = 1;
%     f = 2000;
%     Ps = 30;
%     BW = 800;
    gamma = par.pf_par.gamma;
    beta = par.pf_par.beta;
    f = par.stim_par.fo;
    Ps = par.stim_par.siglev;
    BW = par.stim_par.BW;

    fl = (1-gl)*f-BW;
    fu = (1+gu)*f+BW;
    maxgl = (f-fl)/f;
    maxgu = (fu-f)/f;

    roex_integral = @(p,g) -(1/p)*(2+p*g)*exp(-p*g);

    term_lowertip = roex_integral(pl,maxgl)-roex_integral(pl,gl);
    term_lowertail = roex_integral(t,maxgl)-roex_integral(t,gl);
    term_uppertip = roex_integral(pu,maxgu)-roex_integral(pu,gu);

    term1 = f*((1-10^(w/10))*term_lowertip+(10^(w/10))*term_lowertail+term_uppertip);
    N0 = Ps - 10*log10(term1) - K;
    p = gamma+(1-gamma)*(1+exp((msklev-N0)*beta)).^(-1);

end

function H = roex5_jacobian(x,phi, par)
    % all derivative here are calculated numerically
    % model parameters in phi
    pu = phi(1);
    pl = phi(2);
    t = phi(3);
    logw = phi(4);
    logK = phi(5);
    h = 1e-6;   % this determines the local truncation error
    
    % 1-point numerical derivation (less accurate but fast)
    % ref
    ref = roex5_psycfun(x,[pu,pl,t,logw,logK], par);
    % dpc/dpu
    H(1) = (roex5_psycfun(x,[pu+h,pl,t,logw,logK], par)-ref)/h;
    % dpc/dpl
    H(2) = (roex5_psycfun(x,[pu,pl+h,t,logw,logK], par)-ref)/h;
    % dpc/dt
    H(3) = (roex5_psycfun(x,[pu,pl,t+h,logw,logK], par)-ref)/h;
    % dpc/dlogw
    H(4) = (roex5_psycfun(x,[pu,pl,t,logw+h,logK], par)-ref)/h;
    % dpc/dlogK
    H(5) = (roex5_psycfun(x,[pu,pl,t,logw,logK+h], par)-ref)/h;
end

%%
% service functions for plotting and visualization

function h = iso_density_contour(mu,Sigma,varargin)

    if any(size(Sigma) ~= [2, 2]) || any(size(mu) ~= [1, 2])
        error('The size of the mean or covariance matrix is incorrect.');
    end

    if nargin == 4
        x_lim = varargin{1};
        gradation = varargin{2};
        [X1,X2] = meshgrid(linspace(x_lim(1,1),x_lim(2,1),gradation)',...
            linspace(x_lim(1,2),x_lim(2,2),gradation)');
    elseif nargin == 3
        gradation = 25;
        x_lim = varargin{1};
        [X1,X2] = meshgrid(linspace(x_lim(1,1),x_lim(2,1),gradation)',...
            linspace(x_lim(1,2),x_lim(2,2),gradation)');
    elseif nargin == 2;
        plot_range = 3;
        gradation = 25;
        [X1,X2] = meshgrid(linspace(mu(1)-plot_range*sqrt(Sigma(1,1)),mu(1)+plot_range*sqrt(Sigma(1,1)),gradation)',...
            linspace(mu(2)-plot_range*sqrt(Sigma(2,2)),mu(2)+plot_range*sqrt(Sigma(2,2)),gradation)');
    end
    X = [X1(:) X2(:)];
    p = mvnpdf(X, mu, Sigma);
    [~,h] = contourf(X1,X2,reshape(p,gradation,gradation));
    set(h,'LineStyle','none');

end

%eof